#include "first.h"
#include "Font.h"

#include "Bitmap.h"
#include "Iff.h"
#include "VString.h"

//-----------------------------------------

Font::FontData *Font::fontArray=NULL;

//-----------------------------------------

void Font::install(void)
{
	fontArray=NULL;

	// it is POSSIBLE to load all of the fonts here and refCount them =1.  
	// That would force all fonts to be loaded all the time, preventing disk access during the game.
}

void Font::remove(void)
{
	FontData *temp;

	DEBUG_FATAL(fontArray,("Somebody forgot to release the [%s] font.  RefCount=%d, Accesses=%d.\n",fontArray->filename,fontArray->refCount,fontArray->access));
	while (fontArray)
	{
		temp=fontArray;
		fontArray=fontArray->next;

		delete []temp->filename;
		temp->filename=NULL;

		delete []temp->searchTable;
		temp->searchTable=NULL;

		delete temp->bitmap;
		temp->bitmap=NULL;

		delete temp;
		temp=NULL;
	}
}

//-----------------------------------------

Font::Font(const char *fn,real r,real g,real b,real br,real bg,real bb)
{
	init(fn,r,g,b,br,bg,bb);
}

Font::Font(const char *fn,int r, int g, int b,int br,int bg,int bb)
{
	real red=static_cast<real>(r)/CONST_REAL(255);
	real green=static_cast<real>(g)/CONST_REAL(255);
	real blue=static_cast<real>(b)/CONST_REAL(255);
	real bred=static_cast<real>(br)/CONST_REAL(255);
	real bgreen=static_cast<real>(bg)/CONST_REAL(255);
	real bblue=static_cast<real>(bb)/CONST_REAL(255);
	init(fn,red,green,blue,bred,bgreen,bblue);
}

void Font::init(const char *fn,real fr,real fg,real fb,real br,real bg,real bb)
{
	char bitmapName[128];

	font=fontArray;
	while (font && (stricmp(fn,font->filename) || abs(font->red-fr)>CONST_REAL(0.01) || abs(font->green-fg)>CONST_REAL(0.01) || abs(font->blue-fb)>CONST_REAL(0.01) || abs(font->blendRed-br)>CONST_REAL(0.01) || abs(font->blendGreen-bg)>CONST_REAL(0.01) || abs(font->blendBlue-bb)>CONST_REAL(0.01)))
		font=font->next;

	if (font)
	{
		font->refCount++;
		font->access++;
	}
	else
	{
		Iff iff(fn);

		font=NEW FontData;

		font->next=fontArray;      // link into the head of the static list
		fontArray=font;

		font->filename=strdup(fn);
		font->access=1;
		font->refCount=1;


		iff.seekForm(TAG_FONT);
			iff.seekChunk(TAG_HEAD);
				iff.read_string(bitmapName, sizeof(bitmapName));
				font->spaceCharacter=iff.read_u_int32();  // what is the space character code?
				font->fontWidth=iff.read_u_int32();  // if the font is monospace, the width gets set non-zero.  Zero means proportional.
				font->fontHeight=iff.read_u_int32();
			iff.exitChunk();

			font->bitmap=NEW Bitmap(bitmapName);

			if (iff.seekChunk(TAG_MAP, Iff::FlagOptional))
			{
				font->numChars=iff.chunkLength(sizeof(int32) + sizeof(int32));

				font->searchTable=NEW FontCharData[font->numChars];
				for (int loop=0; loop<font->numChars; loop++)
				{
					font->searchTable[loop].charValue=iff.read_u_int32();
					font->searchTable[loop].frameIndex=iff.read_u_int32();
				}

				// now we bubble sort (ack!) to make certain they are in order
				for (loop=0; loop<font->numChars-1; loop++)
				{
					for (int loop2=loop+1; loop2<font->numChars; loop2++)
					{
						if (font->searchTable[loop].charValue>font->searchTable[loop2].charValue)
						{
							swap(font->searchTable[loop], font->searchTable[loop2]);
						}
					}
				}

				iff.exitChunk();
			}
			else
			{
				font->numChars=font->bitmap->frames();
				font->searchTable=NULL;
			}

		iff.exitForm();
	}
}

Font::~Font(void)
{
	font->refCount--;
	if (!font->refCount)
	{
		delete []font->filename;
		font->filename=NULL;

		delete []font->searchTable;
		font->searchTable=NULL;

		delete font->bitmap;
		font->bitmap=NULL;

		if (fontArray==font)     // deleting the head of the list
		{
			fontArray=fontArray->next;
			delete font;
			font=NULL;
		}
		else                  // find, then wipe out structure in list
		{
			FontData *walker;

			walker=fontArray;
			while (walker->next!=font)
				walker=walker->next;
			if (walker->next!=font)
				Fatal("Font not found in ~Font...refCount must be wrong\n");
			walker->next=font->next;
			delete font;
			font=NULL;
		}
	}
}

int Font::pixelWidth(int chr) const
{
	int frame;

	if (font->fontWidth)
		return font->fontWidth;

	frame=findCharInSearchTable(chr);
	DEBUG_FATAL(frame==-1, ("Character %d not found in font [%s].\n",chr,font->filename));

	return font->bitmap->width(frame)+1;
}

int Font::pixelWidthString(const String &src) const
{
	int width=0;

	for (int loop=0; loop<src.getLength(); loop++)
		width+=pixelWidth(src[loop]);

	return width - (font->fontWidth?0:1);  // make jefftep happy
}

void Font::printString(const String &src,int x0,int y0,int x1,JustificationRule justify,real alpha) const
{
	int x,loop;

	UNREF(alpha);   // until jeff makes his bitmaps take an alpha

	switch (justify)
	{
		case JR_right:
			x=x1-pixelWidthString(src);
			break;
		case JR_centerX:
			x=x0+(x1-x0)/2-pixelWidthString(src)/2;
			break;
		case JR_left:
		default:
			x=x0;
			break;
	}

	for (loop=0; loop < src.getLength(); ++loop)
	{
		if (src[loop]==font->spaceCharacter)  // in the case where the character is a space bar, don't print an invisible poly, just skip the space
			x+=pixelWidth(font->spaceCharacter);
		else
		{
			int frame=findCharInSearchTable(src[loop]);
			font->bitmap->draw(x,y0,frame);
			x+=pixelWidth(src[loop]);
		}
	}
}

bool Font::clipStringByWidth(String *src,String *remainder,int width) const
{
	int lastSpace=-1;
	int stringX=0;

	remainder->setNull();
	for (int loop=0; loop<src->getLength(); loop++)
	{
		if (isspace(src->getChar(loop)) || src->getChar(loop)==font->spaceCharacter)
			lastSpace=loop;

		stringX+=pixelWidth(src->getChar(loop))+1;
		if (stringX>=width)
		{
			DEBUG_FATAL(lastSpace==-1,("Font::clipStringByWidth - could not split string."));
			remainder->append(src->getPointer(lastSpace+1));  // strcpy from the last space to the 
			src->setNull(lastSpace);
			return true;
		}
	}

	return false;
}

int Font::findCharInSearchTable(int chr) const
{
	if (!font->searchTable)
	{
		DEBUG_FATAL(chr >= font->numChars, ("Character %d out of range in font [%s]", chr, font->filename));
		return chr;
	}

	DEBUG_FATAL(chr > font->searchTable[font->numChars-1].charValue, ("Character out of range in font [%s] %d/%d", font->filename,chr,font->searchTable[font->numChars-1].charValue));

//  old linear search through a sorted list... lame
//	for (int loop=0; loop<font->numChars; loop++)
//		if (font->searchTable[loop].charValue==chr)
//			return font->searchTable[loop].frameIndex;

//  new predictive pseudo-binary search
	for (int mid, min = 0, max = font->numChars-1; min <= max; )
	{
#if 0
		mid = (max + min) / 2;
#else
		mid = ((chr - font->searchTable[min].charValue) * (max-min) / (font->searchTable[max].charValue-font->searchTable[min].charValue)) + min;
		DEBUG_FATAL(mid < min || mid > max, ("badness"));
#endif		

		int diff = chr - font->searchTable[mid].charValue;

		if (diff == 0)
			return font->searchTable[mid].frameIndex;
		else
			if (diff < 0)
				max = mid - 1;
			else
				min = mid + 1;
	}

	return -1;  // error, character not found in search table
}
